# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Utilities and helper classes and functions."""
from __future__ import annotations

import contextlib
import io
import logging
import typing

log = logging.getLogger(__name__)

LogLevelCallback = typing.Callable[[str], int]


class LogWriter(io.TextIOBase):
    """File-like object for stream-to-log redirection."""

    def __init__(self,
                 logger: logging.Logger,
                 level: int = logging.INFO,
                 level_cb: typing.Optional[LogLevelCallback] = None) -> None:
        """Initialise the LogWriter."""
        self.logger = logger
        if level_cb is not None:
            self.level_cb = level_cb
        else:
            self.level_cb = lambda line: level

    def detach(self) -> typing.BinaryIO:
        """Detatch the underlying binary buffer."""
        raise io.UnsupportedOperation

    def read(self, size: typing.Optional[int] = -1) -> str:
        """Read from the text stream."""
        raise io.UnsupportedOperation

    def write(self, s: str) -> int:
        """Write the contents of a buffer to the stream."""
        written = 0
        for line in s.rstrip().splitlines():
            level = self.level_cb(line)
            msg = line.strip()
            self.logger.log(level, msg)
            written += len(msg)
        return written

    def redirect_stdout(self) -> typing.ContextManager[typing.Any]:
        """Return a context manager to redirect stdout."""
        return contextlib.redirect_stdout(self)



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""rpkimancer documentation config."""

from __future__ import annotations

import datetime
import importlib.metadata
import pathlib

import docutils

import rpkimancer

import sphinx_readable_theme

import yaml

_dist = importlib.metadata.distribution(rpkimancer.__name__)
_buildtime = datetime.datetime.utcnow()

# -- Project Information

project = _dist.metadata["name"]
author = _dist.metadata["author"]

_from_year = 2021
_to_year = _buildtime.year
if _from_year < _to_year:
    _year_range = f"{_from_year}-{_to_year}"
else:
    _year_range = f"{_from_year}"
copyright = f"{_year_range}, {author}"

release = _dist.version
version = ".".join(release.split(".")[:2])


# -- General configuration

extensions = [
    "sphinx.ext.autodoc",
    "myst_parser",
    "sphinx_multiversion",
]
source_suffix = {
    ".rst": "restructuredtext",
    ".md": "markdown",
}
exclude_patterns = []
templates_path = ['_templates']

# -- HTML output

html_theme = 'readable'
html_theme_path = [sphinx_readable_theme.get_html_theme_path()]
html_static_path = ['_static']
html_sidebars = {
    "**": [
        "localtoc.html",
        "relations.html",
        "versions.html",
        "searchbox.html",
    ],
}

# -- Autodoc configuration

autodoc_member_order = "bysource"
autodoc_typehints = "description"
autodoc_typehints_description_target = "all"

# -- Markdown processing

myst_enable_extensions = [
    "colon_fence",
]

# -- Multiversion processing

smv_branch_whitelist = r"^(?!pyup-|gh-pages).*$"
smv_remote_whitelist = r"^origin$"
smv_prebuild_command = "sphinx-apidoc --separate " \
                                     "--force " \
                                     "--output-dir docs/generated/ " \
                                     "rpkimancer/"

# -- OID registry construction


class OidRegistry(docutils.parsers.rst.Directive):
    """RST directive to render OID registry data."""

    has_content = False
    required_arguments = 0
    option_spec = {
        "path": docutils.parsers.rst.directives.unchanged_required,
    }

    def run(self):
        """Process the directive."""
        registry_path = pathlib.Path(__file__).parent / self.options["path"]
        with open(registry_path) as f:
            data = yaml.safe_load(f)
        root = data["root"]
        arc = data["arc"]

        table = docutils.nodes.table()
        tgroup = docutils.nodes.tgroup(cols=2)
        for w in range(2):
            colspec = docutils.nodes.colspec(colwidth=w + 1)
            tgroup.append(colspec)
        table += tgroup

        thead = docutils.nodes.thead()
        tgroup += thead
        row = docutils.nodes.row()
        for col in ("OID", "Details"):
            entry = docutils.nodes.entry()
            entry += docutils.nodes.paragraph(text=col)
            row += entry
        thead.append(row)

        body = self.render_arc(root, arc)
        tgroup += body

        return [table]

    def render_arc(self, root_oid, arc, body=None):
        """Render the body of the table, recursively."""
        if body is None:
            body = docutils.nodes.tbody()
        for roid, info in arc.items():
            oid = f"{root_oid}.{roid}"
            body.append(self.render_oid_row(oid, info))
            try:
                if (sub_arc := info.get("arc")) is not None:
                    self.render_arc(oid, sub_arc, body=body)
            except AttributeError:
                continue
        return body

    def render_oid_row(self, oid, info):
        """Render the table row for a single OID."""
        row = docutils.nodes.row()

        oid_cell = docutils.nodes.entry()
        oid_para = docutils.nodes.paragraph()
        oid_para += docutils.nodes.literal(text=oid)
        oid_cell += oid_para
        row += oid_cell

        details_cell = docutils.nodes.entry()
        if isinstance(info, str):
            details = [docutils.nodes.paragraph(text=f"{info}")]
        else:
            name = docutils.nodes.paragraph()
            name += docutils.nodes.literal(text=info["name"])
            description = docutils.nodes.paragraph(text=info["description"])
            details = [name, description]
            try:
                for uri in info["refs"]:
                    ref_para = docutils.nodes.paragraph()
                    ref = docutils.nodes.reference(internal=False,
                                                   refuri=uri,
                                                   text=uri)
                    ref_para += ref
                    details.append(ref_para)
            except KeyError:
                pass
        details_cell.extend(details)
        row += details_cell

        return row


def setup(app):
    """Set up local extenstions."""
    app.add_directive("oid-registry", OidRegistry)
    return {"version": "0.1",
            "parallel_read_safe": True,
            "parallel_write_safe": True}



#!/usr/bin/env python
# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""OID registry validator."""

from __future__ import annotations

import json
import logging
import pathlib
import sys

import coloredlogs

import jsonschema

import verboselogs

import yaml

log = verboselogs.VerboseLogger(__name__)

LOG_FMT = "%(levelname)s %(message)s"
LOG_FIELD_STYLES = {"levelname": {"bold": True, "color": "white"}}


def main() -> int:
    """Validate registry data using schema."""
    coloredlogs.install(level=logging.INFO,
                        logger=log,
                        fmt=LOG_FMT,
                        field_styles=LOG_FIELD_STYLES)
    base_path = pathlib.Path(__file__).parent

    schema_path = base_path / "schema.json"
    log.info(f"opening schema '{schema_path}'")
    try:
        with open(schema_path) as f:
            log.info("parsing schema")
            schema = json.load(f)
    except Exception as e:
        log.error(f"failed to read schema: {e}")
        return 2

    data_path = base_path / "registry.yaml"
    log.info(f"opening registry data '{data_path}'")
    try:
        with open(base_path / "registry.yaml") as f:
            log.info("parsing registry data")
            data = yaml.safe_load(f)
    except Exception as e:
        log.error(f"failed to read registry data: {e}")
        return 2

    log.info("trying to get metaschema id")
    metaschema_id = schema.get("$schema")
    if metaschema_id is None:
        log.warning("no '$schema' property found in schema")

    log.info(f"determining validator class for '{metaschema_id}'")
    validator_cls = jsonschema.validators.validator_for(schema, default=None)
    if validator_cls is None:
        log.error("unable to determine validator class")
        return 2

    log.info(f"constructing '{validator_cls.__name__}' validator")
    try:
        validator_cls.check_schema(schema)
        validator = validator_cls(schema)
    except Exception as e:
        log.error(f"failed to construct jsonschema validator: {e}")
        return 2

    log.info("trying to validate registry data")
    errors = sorted(validator.iter_errors(data), key=lambda e: e.path)
    if errors:
        for err in errors:
            log.error(f"validation error:\n{err}")
        return 1

    log.success("no errors found")
    return 0


if __name__ == "__main__":
    rv = main()
    sys.exit(rv)



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""RPKI ROA implementation - RFC6482."""

from __future__ import annotations

import copy
import json
import logging
import typing

from .base import EncapsulatedContentType, SignedObject
from ..asn1.mod import RPKI_ROA
from ..resources import (AFI, IPNetwork, IPNetworkBits, IpResourcesInfo,
                         bitstring_to_net, net_to_bitstring)

log = logging.getLogger(__name__)

RoaNetworkInfo = typing.Tuple[IPNetwork, typing.Optional[int]]


class RouteOriginAttestationContentType(EncapsulatedContentType):
    """encapContentInfo for RPKI ROAs - RFC6482."""

    asn1_definition = RPKI_ROA.ct_routeOriginAuthz
    file_ext = "roa"
    as_resources = None

    def __init__(self, *,
                 version: int = 0,
                 as_id: int,
                 ip_address_blocks: typing.List[RoaNetworkInfo]) -> None:
        """Initialise the encapContentInfo."""
        log.info(f"preparing data for {self}")
        entry_type = typing.Dict[str, typing.Union[IPNetworkBits, int]]

        def address_entry(network: IPNetwork,
                          maxlen: typing.Optional[int] = None) -> entry_type:
            entry: entry_type = {"address": net_to_bitstring(network)}
            if maxlen is not None:
                entry["maxLength"] = maxlen
            return entry

        address_blocks = [{"addressFamily": AFI[network.version],
                           "addresses": [address_entry(network, maxlen)]}
                          for network, maxlen in ip_address_blocks]
        data = {"version": version,
                "asID": as_id,
                "ipAddrBlocks": address_blocks}
        super().__init__(data)
        self._ip_resources = [network for network, _ in ip_address_blocks]

    @property
    def ip_resources(self) -> IpResourcesInfo:
        """Get the IP Address Resources covered by this ROA."""
        return self._ip_resources

    def to_txt(self) -> str:
        """Get default text serialization."""
        return self.to_json()

    def to_json(self) -> str:
        """Serialize as JSON."""
        data = copy.deepcopy(self.content_data)
        afi_bytes_version_map = {v: k for k, v in AFI.items()}
        for i, addr_block in enumerate(self.content_data["ipAddrBlocks"]):
            data_addr_block = data["ipAddrBlocks"][i]
            version = afi_bytes_version_map[addr_block["addressFamily"]]
            data_addr_block["addressFamily"] = f"ipv{version}"
            for j, addr in enumerate(addr_block["addresses"]):
                data_addr = data_addr_block["addresses"][j]
                network = bitstring_to_net(addr["address"], version)
                data_addr["address"] = str(network)
        return json.dumps(data, indent=2)


class RouteOriginAttestation(SignedObject[RouteOriginAttestationContentType]):
    """CMS ASN.1 ContentInfo for RPKI ROAs."""



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""RPKI Manifest implementation - RFC6486."""

from __future__ import annotations

import datetime
import logging
import typing

from .base import EncapsulatedContentType, SignedObject
from ..algorithms import SHA256
from ..asn1.mod import RPKIManifest
from ..resources import INHERIT_AS, INHERIT_IPV4, INHERIT_IPV6

log = logging.getLogger(__name__)

GeneralizedTimeInfo = typing.Tuple[typing.Optional[str], ...]
HashInfo = typing.Tuple[int, int]
FileListInfo = typing.List[typing.Tuple[str, bytes]]


class RpkiManifestContentType(EncapsulatedContentType):
    """encapContentInfo for RPKI Manifests - RFC6486."""

    asn1_definition = RPKIManifest.ct_rpkiManifest
    file_ext = "mft"
    as_resources = INHERIT_AS
    ip_resources: typing.Final = (INHERIT_IPV4, INHERIT_IPV6)

    def __init__(self, *,
                 version: int = 0,
                 manifest_number: int = 0,
                 this_update: datetime.datetime,
                 next_update: datetime.datetime,
                 file_list: FileListInfo) -> None:
        """Initialise the encapContentInfo."""
        log.info(f"preparing data for {self}")
        data = {"version": version,
                "manifestNumber": manifest_number,
                "thisUpdate": self.generalized_time(this_update),
                "nextUpdate": self.generalized_time(next_update),
                "fileHashAlg": SHA256,
                "fileList": [{"file": f[0],
                              "hash": self.hash_bitstring(f[1])}
                             for f in file_list]}
        super().__init__(data)

    @staticmethod
    def generalized_time(timestamp: datetime.datetime) -> GeneralizedTimeInfo:
        """Construct ASN.1 GeneralizedTime data from python datetime."""
        return tuple(f"{t:02}" for t in timestamp.timetuple()[:4]) + \
               tuple(None for _ in range(4))  # noqa: E127

    def hash_bitstring(self, contents: bytes) -> HashInfo:
        """Construct ASN.1 BIT STRING of a hash over file contents."""
        digest = self.digest_algorithm(contents).digest()  # type: ignore[call-arg, misc] # noqa: E501
        hash_bits = int.from_bytes(digest, "big")
        hash_len = len(digest) * 8
        return (hash_bits, hash_len)


class RpkiManifest(SignedObject[RpkiManifestContentType]):
    """CMS ASN.1 ContentInfo for RPKI Manifests."""



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""X.509 ASN.1 Types - RFC5912."""

from __future__ import annotations

import logging

from ..asn1 import Interface, append_info_object_set
from ..asn1.mod import PKIX1Explicit_2009
from ..asn1.types import ASN1Class

log = logging.getLogger(__name__)


class Certificate(Interface):
    """X.509 ASN.1 Certificate type - RFC5912."""

    content_syntax = PKIX1Explicit_2009.Certificate

    @property
    def subject_public_key_info(self) -> SubjectPublicKeyInfo:
        """Get the subjectPublicKeyInfo of the Certificate."""
        log.info(f"trying to get subjectPublicKeyInfo data from {self}")
        with self.constructed() as instance:
            data = instance.get_val_at(["toBeSigned", "subjectPublicKeyInfo"])
        return SubjectPublicKeyInfo(data)

    @classmethod
    def register_ext_type(cls, ext_type: ASN1Class) -> None:
        """Add EXTENSION instance to extnID constraint set."""
        tbs_cert = cls.content_syntax.get_internals()["cont"]["toBeSigned"]
        extns = tbs_cert.get_internals()["cont"]["extensions"]
        extn_item = extns.get_internals()["cont"]
        extn_item_id = extn_item.get_internals()["cont"]["extnID"]
        extn_set = extn_item_id.get_const()["tab"]
        append_info_object_set(extn_set, ext_type)


class SubjectPublicKeyInfo(Interface):
    """X.509 ASN.1 SubjectPublicKeyInfo type - RFC5912."""

    content_syntax = PKIX1Explicit_2009.SubjectPublicKeyInfo



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""RPKI Foo Object."""

from __future__ import annotations

import logging

from rpkimancer.asn1.mod import RpkiFoo
from rpkimancer.resources import INHERIT_AS, INHERIT_IPV4, INHERIT_IPV6
from rpkimancer.sigobj.base import EncapsulatedContentType, SignedObject

log = logging.getLogger(__name__)


class FooObjectContentType(EncapsulatedContentType):
    """encapContentInfo for RPKI Foo Objects."""

    asn1_definition = RpkiFoo.ct_rpkiFooObject
    file_ext = "foo"
    as_resources = INHERIT_AS
    ip_resources = [INHERIT_IPV4, INHERIT_IPV6]

    def __init__(self, content: str) -> None:
        """Initialise the encapContentInfo."""
        data = content.encode()
        super().__init__(data)


class FooObject(SignedObject[FooObjectContentType]):
    """CMS ASN.1 ContentInfo for RPKI Foo Objects."""



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Base classes for RPKI Signed Object implementations - RFC6488."""

from __future__ import annotations

import logging
import os
import typing

from ..algorithms import DIGEST_ALGORITHMS, SHA256
from ..asn1.mod import PKIXAlgs_2009
from ..cert import EECertificate
from ..cms import (ContentInfo,
                   ContentType,
                   EncapsulatedContentInfo,
                   SignedAttributes,
                   SignedData)
from ..resources import AsResourcesInfo, IpResourcesInfo

if typing.TYPE_CHECKING:
    from ..cert import CertificateAuthority

log = logging.getLogger(__name__)

CMS_VERSION: typing.Final = 3

ECT = typing.TypeVar("ECT", bound="EncapsulatedContentType")


class EncapsulatedContentType(ContentType):
    """Base for CONTENT-TYPE instance for RPKI Signed Objects - RFC6488."""

    digest_algorithm = DIGEST_ALGORITHMS[SHA256]
    file_ext: str

    @property
    def as_resources(self) -> typing.Optional[AsResourcesInfo]:
        """Get the AS Number Resources required by this eContent."""
        raise NotImplementedError

    @property
    def ip_resources(self) -> typing.Optional[IpResourcesInfo]:
        """Get the IP Address Resources required by this eContent."""
        raise NotImplementedError

    def digest(self) -> bytes:
        """Calculate the message digest over the DER-encoded eContent."""
        return self.digest_algorithm(self.to_der()).digest()  # type: ignore[call-arg, misc] # noqa: E501

    def signed_attrs(self) -> SignedAttributes:
        """Construct the signedAttrs value from the EncapsulatedContentInfo."""
        return SignedAttributes(content_type=self.content_type,
                                message_digest=self.digest())

    def signed_attrs_digest(self) -> str:
        """Calculate the message digest over the DER-encoded signedAttrs."""
        return self.digest_algorithm(self.signed_attrs().to_der()).hexdigest()  # type: ignore[call-arg, misc] # noqa: E501


class SignedObject(ContentInfo[SignedData], typing.Generic[ECT]):
    """Base CMS ASN.1 ContentInfo for RPKI Signed Objects - RFC5911/RFC6488."""

    econtent_type: typing.Type[ECT]
    ee_cert_cls: typing.Type[EECertificate[ECT]] = EECertificate

    @classmethod
    def __init_subclass__(cls, **kwargs: typing.Any) -> None:
        """Register EncapsulatedContentInfo CONTENT-TYPE for DER encoding."""
        super().__init_subclass__(**kwargs)
        econtent_type = typing.get_args(cls.__orig_bases__[0])[0]  # type: ignore[attr-defined] # noqa: E501
        log.info(f"Adding {econtent_type} to constraining object info set")
        cls.register_econtent_type(SignedData, econtent_type)
        cls.econtent_type = econtent_type

    def __init__(self,
                 issuer: CertificateAuthority,
                 file_name: typing.Optional[str] = None,
                 *args: typing.Any,
                 **kwargs: typing.Any) -> None:
        """Initialise the SignedObject."""
        log.info(f"preparing data for {self}")
        # set object file name
        self._file_name = file_name
        # construct encapContentInfo
        self._econtent = self.econtent_type(*args, **kwargs)
        self._econtent_info = EncapsulatedContentInfo(econtent=self.econtent)
        # construct certificate
        ee_cert = self.ee_cert_cls(signed_object=self,
                                   issuer=issuer,
                                   as_resources=self.econtent.as_resources,
                                   ip_resources=self.econtent.ip_resources)
        # construct signedAttrs
        signed_attrs = self.econtent.signed_attrs()
        # construct signature
        signature = ee_cert.sign_object()

        data = {
            # rfc6488 section 2.1.1
            "version": CMS_VERSION,
            # rfc6488 section 2.1.2 and rfc7935
            "digestAlgorithms": [{"algorithm": SHA256}],
            # rfc6488 section 2.1.3
            "encapContentInfo": self.econtent_info.content_data,
            # rfc6488 section 2.1.4
            "certificates": [
                ("certificate", ee_cert.asn1_cert.content_data),
            ],
            # 'crls' omitted per rfc6488 section 2.1.5
            # rfc6488 section 2.1.6
            "signerInfos": [
                {
                    # rfc6488 section 2.1.6.1
                    "version": CMS_VERSION,
                    # rfc6488 section 2.1.6.2
                    "sid": ("subjectKeyIdentifier", ee_cert.ski_digest),
                    # rfc6488 section 2.1.6.3
                    "digestAlgorithm": {"algorithm": SHA256},
                    # rfc6488 section 2.1.6.4
                    "signedAttrs": signed_attrs.content_data,
                    # rfc6488 section 2.1.6.5 and rfc7935
                    "signatureAlgorithm": {
                        "algorithm": PKIXAlgs_2009.rsaEncryption.get_val(),
                    },
                    # rfc6488 section 2.1.6.6
                    "signature": signature,
                    # 'unsignedAttrs' omitted per rfc6488 section 2.1.6.7
                },
            ],
        }
        super().__init__(content=SignedData(data))

    @property
    def econtent_info(self) -> EncapsulatedContentInfo[ECT]:
        """Get the Signed Object's encapContentInfo."""
        try:
            return self._econtent_info
        except AttributeError:
            eci = EncapsulatedContentInfo[ECT].from_content_info(self)
            self._econtent_info = eci
        return self._econtent_info

    @property
    def econtent(self) -> ECT:
        """Get the Signed Object's eContent."""
        try:
            return self._econtent
        except AttributeError:
            econtent_data = self.econtent_info.econtent_val
            self._econtent = self.econtent_type.from_data(econtent_data)
        return self._econtent

    @property
    def file_name(self) -> str:
        """Construct the file name of the SignedObject."""
        if self._file_name is None:
            return f"{self.econtent.signed_attrs_digest()}.{self.econtent.file_ext}"  # noqa: E501
        else:
            return self._file_name

    def publish(self, *,
                pub_path: str,
                uri_path: str,
                repo_path: str,
                **kwargs: typing.Any) -> None:
        """Publish the SignedObject artifact as a DER file in the PP."""
        with open(os.path.join(pub_path, uri_path, repo_path, self.file_name),
                  "wb") as f:
            f.write(self.to_der())



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""rpkincant conjure plugins for RPKI Foo Objects."""

from __future__ import annotations

import logging
import typing

from rpkimancer.cli import Args
from rpkimancer.cli.conjure import ConjurePlugin, PluginReturn

if typing.TYPE_CHECKING:
    from rpkimancer.cert import CertificateAuthority

log = logging.getLogger(__name__)


class ConjureFoo(ConjurePlugin):
    """rpkincant conjure plugin for RPKI Foo Objects."""

    def init_parser(self) -> None:
        """Set up command line argument parser."""
        self.parser.add_argument("--foo-content",
                                 default="Hello World!",
                                 metavar="<text>",
                                 help="Text string to encode into foo object "
                                      "(default: %(default)s)")

    def run(self,
            parsed_args: Args,
            ca: CertificateAuthority,
            *args: typing.Any,
            **kwargs: typing.Any) -> PluginReturn:
        """Run with the given arguments."""
        # create RSC object
        from .sigobj import FooObject
        log.info("creating foo object")
        FooObject(issuer=ca, content=parsed_args.foo_content)
        return



# PYTHON_ARGCOMPLETE_OK
#
# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""rpkincant CLI tool entrypoint."""

from __future__ import annotations

import importlib.metadata
import logging
import typing

from . import Args, BaseCommand, OptionalArgv, Return

log = logging.getLogger(__name__)


class Cli(BaseCommand):
    """Command-line tools based on the rpkimancer library."""

    from .perceive import Perceive
    from .conjure import Conjure

    default_subcommands = [Conjure, Perceive]

    def init_parser(self) -> None:
        """Set up command line argument parser."""
        subcommands = self.default_subcommands
        log.info("trying to load plugins")
        entry_point_name = "rpkimancer.cli"
        entry_points = importlib.metadata.entry_points()
        for entry_point in entry_points.get(entry_point_name, []):
            cls = entry_point.load()
            if issubclass(cls, BaseCommand):
                subcommands.append(cls)
        subparsers = self.parser.add_subparsers(title="sub-commands",
                                                metavar="<command>",
                                                dest="cmd", required=True)
        for cls in subcommands:
            cls(parent=subparsers)

    def run(self,
            parsed_args: Args,
            *args: typing.Any,
            **kwargs: typing.Any) -> Return:
        """Run with the given arguments."""
        log.info("running sub-command {parsed_args.cmd}")
        rc = parsed_args.run(parsed_args)
        return typing.cast(Return, rc)


def main(argv: OptionalArgv = None) -> Return:
    """Read RPKI artifacts."""
    cli = Cli()
    return cli(argv)



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""RPKI SignedObject implementations."""

from __future__ import annotations

import importlib.metadata
import logging
import typing

from . import gbr, mft, roa
from .base import SignedObject

log = logging.getLogger(__name__)

RpkiGhostbusters = gbr.RpkiGhostbusters
RpkiManifest = mft.RpkiManifest
RouteOriginAttestation = roa.RouteOriginAttestation

S = typing.Type[SignedObject[typing.Any]]


def from_ext(ext: str) -> S:
    """Get a SignedObject by file extension."""
    object_types: typing.List[S] = [RpkiGhostbusters,
                                    RpkiManifest,
                                    RouteOriginAttestation]
    entry_point_name = "rpkimancer.sigobj"
    entry_points = importlib.metadata.entry_points()
    for entry_point in entry_points.get(entry_point_name, []):
        log.info(f"trying to load signed object {entry_point.value}")
        cls = entry_point.load()
        if issubclass(cls, SignedObject):
            object_types.append(typing.cast(S, cls))
        else:
            log.warning(f"signed objects must inherit from {SignedObject}")
    lookup_map = {cls.econtent_type.file_ext: cls
                  for cls in object_types}
    try:
        return lookup_map[ext]
    except KeyError:
        return lookup_map[ext.lstrip(".")]



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Object Identifiers used in RPKI Resource Certificates."""

from __future__ import annotations

import logging
import typing

from cryptography import x509

log = logging.getLogger(__name__)

AIA_CA_ISSUERS_OID: typing.Final = x509.oid.AuthorityInformationAccessOID.CA_ISSUERS  # noqa: E501
SIA_CA_REPOSITORY_OID: typing.Final = x509.oid.SubjectInformationAccessOID.CA_REPOSITORY  # noqa: E501
SIA_MFT_ACCESS_OID: typing.Final = x509.ObjectIdentifier("1.3.6.1.5.5.7.48.10")
SIA_OBJ_ACCESS_OID: typing.Final = x509.ObjectIdentifier("1.3.6.1.5.5.7.48.11")
RPKI_CERT_POLICY_OID: typing.Final = x509.ObjectIdentifier("1.3.6.1.5.5.7.14.2")  # noqa: E501
IP_RESOURCES_OID: typing.Final = x509.ObjectIdentifier("1.3.6.1.5.5.7.1.7")
AS_RESOURCES_OID: typing.Final = x509.ObjectIdentifier("1.3.6.1.5.5.7.1.8")



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Classes implementing CMS ASN.1 types."""

from __future__ import annotations

import logging
import typing

from .algorithms import DigestAlgorithm
from .asn1 import Interface, append_info_object_set
from .asn1.mod import CryptographicMessageSyntax_2009
from .asn1.types import ASN1Class, ASN1Obj, OID

log = logging.getLogger(__name__)

CT = typing.TypeVar("CT", bound="ContentType")
ECT = typing.TypeVar("ECT", bound="ContentType")


class ContentInfo(Interface, typing.Generic[CT]):
    """CMS ASN.1 ContentInfo type - RFC5911."""

    content_syntax = CryptographicMessageSyntax_2009.ContentInfo

    def __init__(self, content: CT) -> None:
        """Initialise the instance from contained ContentData."""
        log.info(f"preparing data for {self}")
        content_type_oid = content.content_type
        content_type_name = content.content_syntax.fullname()
        content_data = content.content_data
        data = {"contentType": content_type_oid,
                "content": (content_type_name, content_data)}
        super().__init__(data)

    @classmethod
    def register_econtent_type(cls,
                               content_type: typing.Type[CT],
                               econtent_type: typing.Type[ECT]) -> None:
        """Add CONTENT-TYPE instance to eContentType constraint set."""
        content = cls.content_syntax.get_internals()["cont"]["content"]
        content_const_set = content.get_const()["tab"].get_val()
        content_type_oid = content_type.content_type
        content_data_inst = list(filter(lambda item: item["id"] == content_type_oid,  # noqa: E501
                                        content_const_set.getv()))[0]["Type"]
        encap_content_info = content_data_inst.get_internals()["cont"]["encapContentInfo"]  # noqa: E501
        econtent_open_type = encap_content_info.get_internals()["cont"]["eContentType"]  # noqa: E501
        econtent_const_set = econtent_open_type.get_const()["tab"]
        append_info_object_set(econtent_const_set,
                               econtent_type.asn1_definition)


class ContentTypeIdDescriptor:
    """Data descriptor for 'content_type' class property."""

    def __get__(self, instance: typing.Optional[CT],
                owner: typing.Type[CT]) -> OID:
        """Get CONTENT-TYPE.&id."""
        return owner.asn1_definition.get_val()["id"]


class ContentTypeSyntaxDescriptor:
    """Data descriptor for 'content_syntax' class property."""

    def __get__(self, instance: typing.Optional[CT],
                owner: typing.Type[CT]) -> ASN1Obj:
        """Get CONTENT-TYPE.&Type."""
        return owner.asn1_definition.get_val()["Type"].get_typeref()


class ContentType(Interface):
    """CMS ASN.1 CONTENT-TYPE instance - RFC5911."""

    asn1_definition: ASN1Class

    content_type = ContentTypeIdDescriptor()
    content_syntax = ContentTypeSyntaxDescriptor()


class SignedData(ContentType):
    """CMS ASN.1 ct-SignedData CONTENT-TYPE instance - RFC5911."""

    asn1_definition = CryptographicMessageSyntax_2009.ct_SignedData


class SignedAttributes(Interface):
    """CMS ASN.1 SignedAttributes type - RFC5911."""

    content_syntax = CryptographicMessageSyntax_2009.SignedAttributes

    def __init__(self, content_type: OID, message_digest: bytes) -> None:
        """Initialise the instance from an eContentType and eContent digest."""
        log.info(f"preparing data for {self}")
        ct_attr_oid = CryptographicMessageSyntax_2009.id_contentType.get_val()
        md_attr_oid = CryptographicMessageSyntax_2009.id_messageDigest.get_val()  # noqa: E501
        data = [
            {
                "attrType": ct_attr_oid,
                "attrValues": [('ContentType', content_type)],
            },
            {
                "attrType": md_attr_oid,
                "attrValues": [('MessageDigest', message_digest)],
            },
        ]
        super().__init__(data)


class EncapsulatedContentInfo(Interface, typing.Generic[CT]):
    """CMS ASN.1 EncapsulatedContentInfo type - RFC5911."""

    content_syntax = CryptographicMessageSyntax_2009.EncapsulatedContentInfo

    digest_algorithm: DigestAlgorithm

    def __init__(self, econtent: CT) -> None:
        """Initialise the instance from contained ContentData."""
        log.info(f"preparing data for {self}")
        data = {"eContentType": econtent.content_type,
                "eContent": econtent.to_der()}
        super().__init__(data)

    @classmethod
    def from_content_info(cls,
                          content_info: ContentInfo[SignedData]) -> EncapsulatedContentInfo[CT]:  # noqa: E501
        """De-encapsulate from ContentInfo instance."""
        val_path = ["content", "SignedData", "encapContentInfo"]
        with content_info.constructed() as instance:
            data = instance.get_val_at(val_path)
        return cls.from_data(data)

    @property
    def econtent_val(self) -> typing.Any:
        """Extract the eContent value."""
        with self.constructed() as instance:
            val = instance.get_val_at(["eContent"])
        if isinstance(val, tuple):
            return val[1][1]
        else:
            return val



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Compile and re-export the provided ASN.1 modules."""
from __future__ import annotations

import glob
import importlib.metadata
import importlib.resources
import logging
import os

import pycrate_asn1c.asnproc as _asn1_compile
import pycrate_asn1c.generator as _asn1_generate

from ...utils import LogWriter

log = logging.getLogger(__name__)
log_writer = LogWriter(log, level=logging.INFO)


def _compile_modules() -> None:
    log.info("Trying to compile ASN.1 modules sources")
    pkg_dir = os.path.dirname(__file__)
    mods_dir = os.path.join(pkg_dir, "modules")
    output_path = os.path.join(pkg_dir, "_mod.py")
    mods = list()
    log.info("reading local distribution modules")
    for path in glob.glob(os.path.join(mods_dir, "**", "*.asn")):
        log.debug(f"Reading {path}")
        with open(path) as f:
            mods.append(f.read())
    log.info("trying to find plugin provided modules")
    entry_point_name = "rpkimancer.asn1.modules"
    entry_points = importlib.metadata.entry_points()
    for entry_point in entry_points.get(entry_point_name, []):
        mod = entry_point.load()
        for item in importlib.resources.contents(mod):
            if not importlib.resources.is_resource(mod, item):
                continue
            if not item.endswith(".asn"):
                continue
            log.info(f"Reading {mod}.{item}")
            with importlib.resources.open_text(mod, item) as f:
                mods.append(f.read())
    with log_writer.redirect_stdout():
        _asn1_compile.compile_text(mods)
        _asn1_generate.PycrateGenerator(dest=output_path)
    log.info("Compilation done")


_compile_modules()
with log_writer.redirect_stdout():
    from ._mod import *  # noqa



# Copyright (c) 2021 Ben Maddison. All rights reserved.
#
# The contents of this file are licensed under the MIT License
# (the "License"); you may not use this file except in compliance with the
# License.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations under
# the License.
"""Number resource ASN.1 type helpers."""

from __future__ import annotations

import ipaddress
import logging
import typing

from .asn1 import Interface
from .asn1.mod import IPAddrAndASCertExtn

log = logging.getLogger(__name__)

AFI = {4: (1).to_bytes(2, "big"),
       6: (2).to_bytes(2, "big")}

Inherit = typing.Literal["INHERIT"]
AfiInfo = typing.Literal[4, 6]

_INHERIT: Inherit = "INHERIT"
INHERIT_AS = _INHERIT
INHERIT_IPV4: typing.Tuple[AfiInfo, Inherit] = (4, _INHERIT)
INHERIT_IPV6: typing.Tuple[AfiInfo, Inherit] = (6, _INHERIT)

IPNetwork = typing.Union[ipaddress.IPv4Network, ipaddress.IPv6Network]
IPNetworkBits = typing.Tuple[int, int]
IPAddressFamilyInfo = typing.Union[typing.Tuple[AfiInfo, Inherit],
                                   IPNetwork]
IpResourcesInfo = typing.Iterable[IPAddressFamilyInfo]
ASIdOrRangeInfo = typing.Union[int, typing.Tuple[int, int]]
AsResourcesInfo = typing.Union[Inherit,
                               typing.Iterable[ASIdOrRangeInfo]]


def net_to_bitstring(network: IPNetwork) -> IPNetworkBits:
    """Convert an IPNetwork to an ASN.1 BIT STRING representation."""
    log.debug(f"converting {network} to rfc3779 bit string")
    netbits = network.prefixlen
    hostbits = network.max_prefixlen - netbits
    value = int(network.network_address) >> hostbits
    return (value, netbits)


def bitstring_to_net(bits: IPNetworkBits, version: int) -> IPNetwork:
    """Convert an ASN.1 BIT STRING representation to an IPNetwork."""
    len_map = {4: ipaddress.IPV4LENGTH, 6: ipaddress.IPV6LENGTH}
    value, netbits = bits
    hostbits = len_map[version] - netbits
    net = ipaddress.ip_network((value << hostbits, netbits))
    return net


class IPAddrBlocks(Interface):
    """ASN.1 IPAddrBlocks type - RFC3779."""

    content_syntax = IPAddrAndASCertExtn.IPAddrBlocks

    def __init__(self, ip_resources: IpResourcesInfo) -> None:
        """Initialise instance from python data."""
        log.info(f"preparing data for {self}")
        net_data_type = typing.Union[Inherit,
                                     typing.Tuple[str, IPNetworkBits]]
        entry_type = typing.Tuple[int, net_data_type]

        def _net_entry(network: IPAddressFamilyInfo) -> entry_type:
            if isinstance(network, (ipaddress.IPv4Network,
                                    ipaddress.IPv6Network)):
                return network.version, ("addressPrefix",
                                         net_to_bitstring(network))
            else:
                return network[0], _INHERIT

        combined_type = typing.Tuple[str,
                                     typing.Union[typing.Literal[0],
                                                  typing.List[net_data_type]]]

        def _combine(entries: typing.List[net_data_type]) -> combined_type:
            if any(entry == _INHERIT for entry in entries):
                return ("inherit", 0)
            else:
                return ("addressesOrRanges", [entry for entry in entries])

        by_afi = {afi_data: [net_data
                             for net_version, net_data
                             in map(_net_entry, ip_resources)
                             if net_version == afi_version]
                  for (afi_version, afi_data) in AFI.items()}
        data = [{"addressFamily": afi, "ipAddressChoice": _combine(entries)}
                for afi, entries in by_afi.items() if entries]
        super().__init__(data)


class ASIdOrRange(Interface):
    """ASN.1 ASIdOrRange type - RFC3779."""

    content_syntax = IPAddrAndASCertExtn.ASIdOrRange

    def __init__(self, a: ASIdOrRangeInfo) -> None:
        """Initialise instance from python data."""
        log.info(f"preparing data for {self}")
        data: typing.Union[typing.Tuple[str, int],
                           typing.Tuple[str, typing.Dict[str, int]]]
        if isinstance(a, int):
            data = ("id", a)
        elif isinstance(a, tuple):
            data = ("range", {"min": a[0], "max": a[1]})
        super().__init__(data)


class ASIdentifiers(Interface):
    """ASN.1 ASIdentifiers type - RFC3779."""

    content_syntax = IPAddrAndASCertExtn.ASIdentifiers

    def __init__(self, as_resources: AsResourcesInfo) -> None:
        """Initialise instance from python data."""
        asnum: typing.Union[typing.Tuple[str, int],
                            typing.Tuple[str, typing.List[typing.Any]]]
        if as_resources == INHERIT_AS:
            asnum = ("inherit", 0)
        elif isinstance(as_resources, list):
            asnum = ("asIdsOrRanges",
                     [ASIdOrRange(a).content_data for a in as_resources])
        else:  # pragma: no cover
            raise ValueError
        data = {"asnum": asnum}
        super().__init__(data)
